#!/usr/bin/perl
# 2011-03-26 eat - added onclick to image input for compatibility with FireFox 4.0 and IE

# This class is used to manage the permits element used within control elements
#
#   <permits>
#     <permit state='' time='' planner='' reason=''
#
#        planner  - UserId of permitted planner
#
#        role     - Value assigned to attribute
#   </permits>

package Permits;
use Carp;
use NBK::Planner::Planner;

#===========================================================================================================
# Scheme Methods

# Create new object

sub new{
  my ($package,$planner,$doc)=@_;
  bless {PLANNER => $planner, DOC => $doc} => $package;
  }

# We get to this update method via a section identifier
#
#    request=update;path=<planPath>:<type>.<element>[(<key>)]...;action=<action>;...
#
# This function is only called when the section identifier locates this COLUMNS element.
#
sub getUpdate{
  my ($self,$action,$upHash)=@_; 
  my $planner=$self->{PLANNER};
  my $msg=''; 
  if($action eq 'Toggle' || $action eq 'Delete'){
    my $state='deleted';
    my @id=$planner->getParams('enable');
    $state='enabled' if($action eq 'Toggle');
    for my $id (@id){
      $$upHash{"$id"}=$state;
      }
    @id=$planner->getParams('disable');
    $state='disabled' if($action eq 'Toggle');
    for my $id (@id){
      $$upHash{"$id"}=$state;
      }
    @id=$planner->getParams('delete');
    for my $id (@id){
      $$upHash{"$id"}='deleted';
      }
    }
  elsif($action eq 'Update' || $action eq 'Add'){
    my $id=$planner->getParam('user');
    my $userObj=$planner->getUser($id);
    my $permitUserName=$userObj->getName();
    my $state=$planner->getParam('state');
    my $xml="<permit user='$id'";
    $xml.=" state='$state'" if($state ne 'enabled');
    my $utime=time();
    my $userId=$planner->getUserId();
    my $reason=$planner->getParam('reason');
    $reason=$planner->encode($reason);
    $xml.=" time='$utime' planner='$userId' reason='$reason'";
    my $permitRole=$planner->getParam('role');
    $xml.=" role='$permitRole'";
    $xml.="/>\n";
    $$upHash{"$id"}=$xml;

    #my $foo=$planner->encode($xml);
    #print("XML:$foo<br>\n");
    }
  return($msg);
  }

# Convert a list document into XML text
#   Note: we are only interested here in the list part
#         the other plan elements are handled by Plan.pm
#
#   request=Update;path=<planPath>:<planName>:<type>.<element>[(<key>)]...;action=<action>;...
#   request=Update;path=<planPath>:<planName>:List.table;action=<action>;...
#
# Parameters obtained by calls to planner->getParam():
#
#   permit*
#
#     action='Toggle' 
#
#       enabled=<rowid>;enabled=<rowid>;disabled=<rowid>;disabled=<rowid>;...
#
#     action='Update'
#
#       row=<rowId>;cells=<columnNum>;cell1=<value>;cell2=<value>;cell3=<value>;...
#
# Update hashes:
#   $upHash - $$upRow{$id}='enabled'|'disabled'|'deleted'|'<...new row XML...>'
#
# Returns string that is XML if it starts with '<', otherwise and error message

sub xml{
  my ($self,$path,$element,$action,$canEdit,$canCntl)=@_;
  #print("Permits.xml called element=$element action=$action<br>\n");
  my $planner=$self->{PLANNER};
  my $permitsDoc=$self->{DOC};  
  my %permit;
  my %upHash;
  my $xml='';
  
  foreach my $permitDoc ($permitsDoc->children('permit')){
    next if(!$permitDoc);
    $permit{$permitDoc->attribute('user')}=$permitDoc;
    }
  # Get update hashes from parameters if there is an update for this element
  if($action && $element=~/^permit/){
    return("User does not have control authority.") if(!$canCntl);
    $msg=$self->getUpdate($action,\%upHash) if($action);
    return($msg) if($msg);
    }
  # <permits>
  $xml.="<permits>\n";
  foreach my $id (sort {$a<=>$b} keys %permit){
    my $update=$upHash{$id};
    my $permitDoc=$permit{$id};
    my $userObj=$planner->getUser($id);
    my $userName=$userObj->getName();
    my $permitRole=$permitDoc->attribute('role');
    if($update=~/^</){
      $xml.=$update;
      $upHash{$id}='';
      $planner->postChange('PlanControlPermitUpdate',$path,"($id,$userName)");
      }
    elsif($update eq 'deleted'){   # note: when a column is deleted corresponding cells will be deleted from table rows
      $planner->postChange('PlanControlPermitDelete',$path,"($id,$userName,$permitRole)");
      }
    elsif($update eq 'enabled' || $update eq 'disabled'){
      if($update eq 'disabled'){
        $xml.="<permit user='$id' state='disabled'";
        $verb='Disable';
        }
      else{
        $xml.="<permit user='$id'";
        $verb='Enable';
        }
      my $elementTime=time();
      my $elementUserId=$planner->getUserId();
      my $elementReason=$planner->getParam('reason');
      $xml.=" time='$elementTime' planner='$elementUserId' reason='$elementReason'"; 
      $xml.=" role='$permitRole'";
      $xml.="/>\n";
      $planner->postChange("PlanControlPermit$update",$path,"($id,$userName,$permitRole)");
      }
    else{$xml.=$planner->getElementXml('permit',$permitDoc,'user,state,time,planner,reason,role',"/>\n");}
    }
  # handle additions
  foreach my $id (sort {$a<=>$b} keys %upHash){
    if($upHash{$id}=~/^</){
      $xml.=$upHash{$id};
      $permitRole='';
      $permitRole=$1 if($upHash{$id}=~/role='(.*?)'/);
      $planner->postChange('PlanControlPermitAdd',$path,"($id,$permitRole)");
      }
    }
  $xml.="</permits>\n";
  return($xml);  
  }


# Return HTML representation
#
#   Path  - portion of section consumed by parent element
#           This is required for constructing URL's to get
#           back for update. 
#
#   Action  Element
#   ------  --------
#
#   'View'  columns
#           column(<colId>)
#
#   'Edit'  columns
#           column(<colId>)
#   
sub html{
  my ($self,$path,$element,$action,$canEdit,$canCntl)=@_;
  #print("Permits.html called path=$path element=$element action=$action<br>\n");
  my $planner=$self->{PLANNER};
  my $permitsDoc=$self->{DOC};
  $permitId=$1 if($element=~/^permit\((\d+)\)$/);
  $action='Display' if(!$action);  # default action
  #$element='permits' if(!$element);
  my $html='';
  my %permit;
  my $num=0;
  foreach my $permitDoc ($permitsDoc->children('permit')){
    next if(!$permitDoc);
    $permit{$permitDoc->attribute('user')}=$permitDoc;
    $num++;
    }
  my $editable=1 if($action eq 'Edit' && $canCntl && (!$element || $element=~/^permit/));
  # permits
  my $request='Edit';
  $request='Display' if($action eq 'Edit');
  $html.="<table width='100%' border=1 cellspacing=0 cellpadding=0><tr><th><a href='Planner.cgi?request=$request;path=$path'><img src='planControl.gif' border=0 title='$request permissions'></a> Permissions</th></tr><tr><td>\n";
  if($editable){  # If edit requested and authorized
    $html.="<form style='margin-bottom:0;' name=Permits action='Planner.cgi' method=post>\n";
    $html.="<input type='hidden' name='request' value='Update'/>\n";
    $html.="<input type='hidden' name='action' value=''/>\n";
    $html.="<input type='hidden' name='path' value='$path.permits'/>\n";
    my ($permitDoc,$permitUser,$permitRole);
    if($permitId){
      $permitDoc=$permit{$permitId};
      if($permitDoc){
        $permitState=$permitDoc->attribute('state');
        $permitUser=$permitDoc->attribute('user');
        $permitRole=$permitDoc->attribute('role');
        }
      }
    $html.="<table width='100%' class='fill'><tr><th>Permit</th><th>";
    $html.="<select name='state'>\n";
    foreach my $state ('enabled','disabled'){
      my $selected='';
      $selected=' SELECTED' if($state eq $permitState);
      $html.="<option value='$state'$selected>$state</option>\n";
      }
    $html.="</select>\n";
    $html.="&nbsp; <input type=image src='planAdd.gif' title='Create new permit' name='image' onclick=\"this.form.action.value='Add'\"/>\n";
    $html.="&nbsp; <input type=image src='planUpdate.gif' title='Write changes to selected permit' name='image' onclick=\"this.form.action.value='Update'\">\n" if($permitDoc);
    $html.="</td></tr>\n";
    $html.="<tr><th width='5%'>Parameter</th><th>Value</th><tr>\n";
    # id,name,value,type,for,format,audit,size,default,term,state,comment,tip,title
    $html.="<tr><td><b>User</b></td><td>";
    if($permitDoc){
      $html.="<input type=hidden name=user value='$permitId'>\n" if($permitId);
      my $userObj=$planner->getUser($permitId);
      my $userName=$userObj->getName();
      $html.=$userName;
      }
    else{
      $html.=$planner->getPlannerMenu();
      $html.=" User granted permission to perform the identified role.</td><tr>";
      }
    $html.="<tr><td><b>Role</b></td><td><select name=role>\n";
    foreach my $option ('planner','admin','owner'){
      my $selected='';
      $selected=' SELECTED' if($option eq $permitRole);
      $html.="<option value='$option'$selected>$option</option>\n";
      }
    $html.="</select> Permission granted.  Use [planner] is recommended in most cases.</td><tr>";
    $html.="<tr><th>Reason</th><th colspan=2><input type=text name=reason size=80> Change request number and/or short reason for change.</th></tr>\n";
    $html.="</table>\n";
    }
  $html.="<table width='100%' border=0 cellpadding=3 cellspacing=1>\n";
  $html.="<tr><th></th>";  
  if($editable){
    $html.="<th width='37'>";
    $html.="<input type=image src='planToggle.gif' name='image' onclick=\"this.form.action.value='Toggle'\" title='Toggle state of selected permit'/>\n";
    $html.="<input type=image src='planCut.gif' name='image' onclick=\"this.form.action.value='Delete'\" title='Delete selected permit'/></th>\n";
    }
  else{$html.="<th width='19'></th>";}
  $html.="<th>Name</th><th>Role</th><th>Email</th><th>Revised</th><th>Planner</th><th>Reason</th></tr>\n";
  $row=0;
  foreach my $id (sort keys %permit){
    my $permitDoc=$permit{$id};
    my $state=$permit{$id}->attribute('state');
    my $permitUserId=$permit{$id}->attribute('user');
    my $userObj=$planner->getUser($permitUserId);
    my $permitUserName=$userObj->getName();
    $permitUserName=$permitUserId if(!$permitUserName);
    my $permitUserEmail=$userObj->getEmail();
    my $permitRole=$permit{$id}->attribute('role');
    $permitRole='planner' if(!$permitRole);
    my $elementTime=$permitDoc->attribute('time');
    my $elementDate=$planner->getUserDate($elementTime) if($elementTime);
    my $elementUserId=$permitDoc->attribute('planner');
    my $elementUserName='';
    if($elementUserId){
      my $userObj=$planner->getUser($elementUserId);
      $elementUserName=$userObj->getName();
      }
    #my $elementUserName=$planner->getUserName($elementUserId) if($elementUserId);
    my $elementReason=$permitDoc->attribute('reason');
    $class='even';
    $class='odd' if($row%2);
    $class='' if($permitUserId eq $permitId);
    $row++;
    $html.="<tr class=$class><td>$row</td>\n";
    my $linkRequest='Display';
    my $linkTitle='Select to highlight this entry';
    if($canCntl){
      $linkRequest='Edit';
      $linkTitle='Select to edit this entry';
      }
    if($state eq 'deleted'){
      $fontOn="<font color='#b0b0b0'>";  # should use style instead
      $fontOff="</font>";
      $stateImage='deleted.gif';
      $checkName='enable';
      }
    elsif($state eq 'disabled'){
      $fontOn="<font color='#b0b0b0'>";
      $fontOff="</font>";
      $stateImage='planDisabled.gif';
      $checkName='enable';
      }
    else{
      $fontOn="";
      $fontOff="";
      $stateImage='planEnabled.gif';
      $checkName='disable';
      }

    $html.="<td><a href='Planner.cgi?request=$linkRequest;path=$path.permit($id)'><img src='$stateImage' border=0 title='$linkTitle'></a>";
    $html.="<input type=checkbox name='$checkName' value='$permitUserId'>" if($editable);
    $html.="</td>";
    foreach my $value ($permitUserName,$permitRole,$permitUserEmail){
       $html.="<td>$fontOn$value$fontOff</td>";
       }
    $html.="<td class=fill>$elementDate</td><td class=fill>$elementUserName</td><td class=fill>$elementReason</td></tr>";
    }
  $html.="</table>";
  $html.="</form>" if($editable);
  $html.="</td></tr></table>\n";
  return($html);
  }

sub nb{
  my ($self)=@_;
  my $planner=$self->{PLANNER};
  my $permitsDoc=$self->{DOC};
  my $nb="# Permits:\n#\n";
  return($nb) if(!$permitsDoc);
  foreach my $permitDoc ($permitsDoc->children('permit')){
    my $permitUser=$permitDoc->attribute('user');
    my $permitRole=$permitDoc->attribute('role');
    $permitRole='planner' if(!$permitRole);
    my $adminObj=$planner->getUser($permitUser);
    if($adminObj){
      $userName=$adminObj->getName();
      $userEmail=$adminObj->getEmail();
      $nb.="#   $userName ($permitRole) <$userEmail>\n";
      }
    else{$nb.="#   $userId ($permitRole)\n";}
    }
  return($nb);
  }

1;
