#!/usr/bin/perl
#
# 2011-03-13 eat 0.8.5 - modified extract to support "_state" attribute
#            This same approach can be used for other row attributes if needed
# 2011-03-26 eat - added onclick to image input for compatibility with FireFox 4.0 and IE
# 2011-10-23 eat - darkened disabled row text to make more readable
# 2011-11-02 eat - fixed defect in dropdown menu display---was reverting to first menu item on updates

package Table;
use Carp;
use NBK::Planner::Planner;
use NBK::Planner::Attributes;
use NBK::Planner::Options;

#===========================================================================================================
# Scheme Methods

# Create new object

sub new{
  my ($package,$planner,$doc)=@_;
  bless {PLANNER => $planner, DOC => $doc} => $package;
  }

# We get to this update method via a path identifier
#
#    request=update;path=<planPath>:<type>.<element>[(<key>)]...;action=<action>;...
#
# This function on called when this table.row* element is specified.
#
sub getUpdate{
  my ($self,$action,$upRow,$columnHash)=@_; 
  my $planner=$self->{PLANNER};
  my $msg=''; 
  if($action eq 'MoveUp' || $action eq 'MoveDown'){
    my @id=$planner->getParams('enable');
    foreach my $id (@id){
      $$upRow{$id}='selected';
      } 
    @id=$planner->getParams('disable');
    foreach my $id (@id){
      $$upRow{$id}='selected';
      }
    }
  elsif($action eq 'Toggle' || $action eq 'Delete'){
    my $state='deleted';
    my @id=$planner->getParams('enable');
    $state='enabled' if($action eq 'Toggle');
    for my $id (@id){
      $$upRow{$id}=$state;
      }
    @id=$planner->getParams('disable');
    $state='disabled' if($action eq 'Toggle');
    for my $id (@id){
      $$upRow{$id}=$state;
      }
    }
  elsif($action eq 'Update' || $action eq 'Add'){
    my $rowId;
    if($action eq 'Update'){$rowId=$planner->getParam('row');}
    else{$rowId='-1';}
    my $rowState=$planner->getParam('state');
    my $xml="<row id='$rowId'";
    $xml.=" state='$rowState'" if($rowState ne 'enabled');
    my $utime=time();
    my $userId=$planner->getUserId();
    my $rowReason=$planner->getParam('reason');
    $rowReason=$planner->encode($rowReason);
    $xml.=" time='$utime' user='$userId' reason='$rowReason'>";
    #my $key='(';
    foreach my $columnId (sort {$a<=>$b} keys %$columnHash){
      my $value=$planner->getParam("cell$columnId");
      if($value){
        my $format=$$columnHash{$columnId}->attribute('format');
        if($format eq 'ip'){
          $value=sprintf("%3.3d.%3.3d.%3.3d.%3.3d",$1,$2,$3,$4) if($value=~/^(\d+)\.(\d+).(\d+).(\d+)$/);
          }
        $value=$planner->encode($value);
        my $for=$$columnHash{$columnId}->attribute('for');
        #$key.=$value.',' if(!$for || $for eq 'arg');
        $xml.="<cell id='$columnId'>$value</cell>";
        }
      }

    #chop($key);
    #$key.=')';
    $xml.="</row>\n";
    $$upRow{"$rowId"}=$xml;
    }
  return($msg);
  }

# Get row key - see nb{} method for sample code
#   This produces a string that identifies a table row using the for=arg attributes
sub getRowKey{
  my ($self,$rowDoc,$columnHash)=@_;
  my $key='(';
  my %cellValue;
  foreach my $cellDoc ($rowDoc->children('cell')){
    $cellValue{$cellDoc->attribute('id')}=$cellDoc->value;
    }
  foreach my $columnId (sort {$a<=>$b} keys %$columnHash){
    my $for=$$columnHash{$columnId}->attribute('for');
    $key.=$cellValue{$columnId}.',' if(!$for || $for eq 'arg');
    }
  chop($key);
  $key.=')'; 
  return($key);
  }

# Convert a list document into XML text
#   Note: we are only interested here in the list part
#         the other plan elements are handled by Plan.pm
#
#   request=Update;path=<planPath>:<type>.<element>[(<key>)]...;action=<action>;...
#   request=Update;path=<planPath>:List.table;action=<action>;...
#
# Parameters obtained by calls to planner->getParam():
#
#   List.table.row*
#
#     action='Toggle' 
#
#       enabled=<rowid>;enabled=<rowid>;disabled=<rowid>;disabled=<rowid>;...
#
#     action='Update'
#
#       row=<rowId>;cells=<columnNum>;cell1=<value>;cell2=<value>;cell3=<value>;...
#
# Update hashes:
#   $upRow - $$upRow{$id}='enabled'|'disabled'|'deleted'|'<...new row XML...>'
#
# Returns string that is XML if it starts with '<', otherwise and error message
sub xml{
  my ($self,$path,$element,$action,$canEdit,$canCntl)=@_;
  my $planner=$self->{PLANNER};
  my $tableDoc=$self->{DOC};
  my %column;
  my %row;
  my %cell;
  my $attributesDoc=$tableDoc->child('attributes');
  $attributesDoc=$tableDoc->child('columns') if(!$attributesDoc);
  my %upRow;
  
  return("User does not have edit authority.") if(!$canEdit);

  my $change=$planner->getParam('change'); # get user's description of change
  $planner->postChange('Edit',$path,$change);

  my $childElement=$1 if($element=~/^.*?\.(.*)$/);
  # support new and old element names
  foreach my $elementName ('attribute','column'){
    foreach my $column ($attributesDoc->children($elementName)){
      next if(!$column);
      $column{$column->attribute('id')}=$column;
      }
    }
  # Get update hashes from parameters if there is an update for this element
  if($action && ($element=~/^table.row/ || $element=~/^table.entry/)){
    $msg=$self->getUpdate($action,\%upRow,\%column) if($action);
    #return($msg) if($msg);
    }
  my $xml='';
  # <table ...
  $xml.=$planner->getElementXml('table',$tableDoc,'name,state,comment',">\n");

  # <attributes>
  
  my $childElement=$1 if($element=~/^table.(.*)$/);
  my $attributesObj=new Attributes($planner,$attributesDoc);
  my ($attributesXml,$cellMap)=$attributesObj->xml("${path}.table",$childElement,$action,$canEdit,$canCntl);
  return($attributesXml) if(!($attributesXml=~/^</));
  $xml.=$attributesXml;

  my %cellMap;
  if($cellMap){
    foreach my $idmap (split(";",$cellMap)){
      my ($wasId,$isId)=split("=",$idmap);
      $cellMap{$wasId}=$isId;
      }
    }
  # <rows>
  $xml.="<rows>\n";
  my $rows=$tableDoc->child('rows');
  foreach my $row ($rows->children('row')){
    next if(!$row);
    $row{$row->attribute('id')}=$row;
    }
  if($action eq 'MoveUp' || $action eq 'MoveDown'){
    my @row;
    my %tab;
    my $target;
    my @mid;
    my @mrow;
    my $target;
    my @mid;
    if($action eq 'MoveUp'){($target,@mid)=sort {$a<=>$b} keys %upRow;}
    else{($target,@mid)=sort {$b<=>$a} keys %upRow;}
    if($target && @mid){
      foreach my $id (@mid){
        push(@mrow,$row{$id});
        delete $row{$id};
        }
      my $i=0;
      foreach my $id (sort {$a<=>$b} keys %row){
        $i++;
        if($upRow{$id}){
          if($id==$target){
            if($action eq 'MoveUp'){
              push(@row,@mrow);
              push(@row,$row{$id});
              }
            else{
              push(@row,$row{$id});
              push(@row,@mrow);
              }
            }  
          # else ignore now because it is a moving row
          } 
        else{push(@row,$row{$id});} 
        }
      my $i=0;
      foreach my $row (@row){
        $i++;
        $tab{$i}=$row;
        }
      %row=%tab;
      }
    }
  my $newId=0;
  foreach my $id (sort {$a<=>$b} keys %row){
    my $update=$upRow{$id};
    $update='order' if($action=~"^Move");
    if($update=~/^</){
      $xml.=$update;
      #delete $upRow{$id};  # delete so we know it isn't an addition later
      $upRow{$id}=''; # a different way
      $key='(';
      $key.=$1 if($update=~/<cell .*?>(.*?)</);
      $key.=',...)';
      #$planner->postChange('PlanTableRowUpdate',$path,"$key);
      $planner->postChange('PlanTableRowUpdate',$path,"$id");
      }
    elsif($update eq 'deleted'){
      #$planner->postChange('PlanTableRowDelete',$path,$self->getRowKey($row{$id},\%column));
      $planner->postChange('PlanTableRowDelete',$path,$id);
      }
    else{
      if($update eq 'disabled' || $update eq 'enabled'){
        if($update eq 'disabled'){
          $xml.="<row id='$id' state='disabled'";
          $verb='Disable';
          }
        else{
          $xml.="<row id='$id'";
          $verb='Enable';
          }
        my $rowTime=time();
        my $rowUser=$planner->getUserId();
        my $rowReason=$planner->getParam('reason');
        $rowReason=$planner->encode($rowReason);
        $xml.=" time='$rowTime' user='$rowUser' reason='$rowReason'>"; 
        #$planner->postChange('PlanTableRow'.$verb,$path,$self->getRowKey($row{$id},\%column));
        $planner->postChange('PlanTableRow'.$verb,$path,$id);
        }
      elsif($update eq 'order'){
        my $row=$row{$id};
        my $rowState=$row->attribute('state');
        $rowState=" state='$rowState'" if($rowState);
        my $rowTime=$row->attribute('time');
        $rowTime=" time='$rowTime'" if($rowTime); 
        my $rowUser=$row->attribute('user');
        $rowUser=" user='$rowUser'" if($rowUser); 
        my $rowReason=$row->attribute('reason');
        $rowReason=$planner->encode($rowReason);
        $rowReason=" reason='$rowReason'" if($rowReason); 
        $xml.="<row id='$id'$rowState$rowTime$rowUser$rowReason>"; 
        }
      else{$xml.=$planner->getElementXml('row',$row{$id},'id,state,time,user,reason','>');}

      undef %cell;
      my %cell;
      foreach my $cell ($row{$id}->children('cell')){
        $cell{$cell->attribute('id')}=$cell;
        }
      foreach my $column (sort {$a<=>$b} keys %cell){
        my $id;
        $value=$cell{$column}->value;
        if($value ne $column{$column}->attribute('default')){  # don't waste space storing default values
          $id=$cell{$column}->attribute('id');
          $id=$cellMap{$id} if($cellMap);
          $xml.="<cell id='".$id."'>"; 
          $xml.=$planner->encode($value);
          $xml.='</cell>';
          }
        }
      $xml.="</row>\n";
      $newId=$id;
      }
    }
  # handle additions
  foreach my $id (sort {$a<=>$b} keys %upRow){
    if($id<0 && $upRow{$id}=~/^</){
      my $rowStr=$upRow{$id};
      $newId++;
      $rowStr=~s/^<row id='-\d+'/<row id='$newId'/g;
      $xml.=$rowStr;
      $key='(';
      $key.=$1 if($rowStr=~/<cell .*?>(.*?)</);
      $key.=',...)';
      $planner->postChange('PlanTableRowAdd',$path,$key);
      }
    }
  $xml.="</rows>\n";
  $xml.="</table>\n";
  $planner->postChange('PlanTableOrder',$path,$action) if($action=~/^Move/);
  return($xml);  
  }


# Return HTML representation
#
#   Action  Element
#   ------  --------
#
#   'View'  table
#           table.colums
#           table.rows
#           table.entries
#
#   'Edit'  table.col(<colId>)
#           table.row(<rowId>)
#           table.entry(<rowId>)
#   
# 2010-11-02 eat - including "table.entries" element
#            When this element is specified, a "vertical" option
#            is set to display each row as a block within a
#            two column table (attribute,value).  This enables
#            long rows to fit on a display better.
#            
sub html{
  my ($self,$path,$element,$action,$canEdit,$canCntl)=@_;
  my $planner=$self->{PLANNER};
  my $tableDoc=$self->{DOC};
  my $rowId='';   # row we are focused on
  my $vertical=0;
  if($element=~/^table\.entries$/){
    $vertical=1;
    $element='table.rows';
    }
  elsif($element=~/^table\.entry\((\d+)\)$/){
    $element="table.row($1)";
    $vertical=1;
    }
  elsif($element eq 'table.entry'){
    $element='table.row';
    $vertical=1;
    }
  elsif(!($element=~/^table\./)){
    $element='';
    $action='Display';
    }
  $rowId=$1 if($element=~/^table\.row\((\d+)\)$/);
  $path=$self->{PATH} if(!$path);  # default path
  $action='Display' if(!$action);  # default action
  $element='table.rows' if(!$element);
  return("Logic error - Table class sent element not addressing Table") if(!($element=~/^table\./));
  my $html='';
  $html.="<table width='100%' border=1 cellspacing=0 cellpadding=0><tr><th>\n";
  my $attrRequest='Edit';
  my $attrRequest2='Edit';
  my $attrElement='table.columns';
  my $attrElement2='table.options';
  #if($action eq 'Edit' && $element=~/^table.(col|opt)/){
  if($action eq 'Edit'){
    if($element=~/^table.column/){
      $attrRequest='Display';
      $attrElement='table.rows';
      }
    if($element=~/^table.options/){
      $attrRequest2='Display';
      $attrElement2='table.rows';
      } 
    }
  if($element=~/^table.options/){
    $html.="<a href='Planner.cgi?request=$attrRequest2;path=$path.$attrElement2'><image src='planOptions.gif' border=0 title='Open options'></a>&nbsp;";
    $html.="<a href='Planner.cgi?request=$attrRequest;path=$path.$attrElement'><img src='planTable.gif' border=0 title='Open attributes'></a>&nbsp;";
    }
  else{
    $html.="<a href='Planner.cgi?request=$attrRequest;path=$path.$attrElement'><img src='planTable.gif' border=0 title='Open attributes'></a>&nbsp;";
    $html.="<a href='Planner.cgi?request=$attrRequest2;path=$path.$attrElement2'><image src='planOptions.gif' border=0 title='Open options'></a>&nbsp;";
    }
  $html.="Table Attributes</th></tr><tr><td>\n";
  my %column;
  my %row;
  my $rowsDoc=$tableDoc->child('rows');
  my $attributesDoc=$tableDoc->child('attributes');
  $attributesDoc=$tableDoc->child('columns') if(!$attributesDoc);
  my $colnum=0;
  my $haveOptions=0;
  # support both new and old element names
  foreach my $elementName ('attribute','column'){
    foreach my $column ($attributesDoc->children($elementName)){
      next if(!$column);
      my $columnState=$column->attribute('state');
      my $columnFor=$column->attribute('for');
      $haveOptions=1 if($columnFor eq 'opt');
      $column{$column->attribute('id')}=$column if($columnFor ne 'opt' && (!$columnState || $columnState eq 'enabled'));
      $colnum++;
      }
    }
  #if(($haveOptions && $element=~/^table.row/) || $element=~/^table.options/){
  if($element=~/^table.row/ || $element=~/^table.options/){
    my $optionsObj=new Options($planner,$tableDoc);
    my $optionsAction=$action;
    $optionsAction='Display' if($element=~/^table.row/);
    my $optionsHtml=$optionsObj->html("$path.table",'options',$optionsAction,$canEdit,$canCntl);
    #my $optionsHtml=$optionsObj->html("$path.table",'',$optionsAction,$canEdit,$canCntl);
    $html.=$optionsHtml;
    }
  if($element=~/^table.options/){
    $element='table.rows';
    $action='Display';
    }
  if($colnum==0){  # display attributes until we have some
    $element='table.columns';
    $action='Edit' if($action eq 'Display' && $canEdit);
    }
  my $tableSwitchPath;
  my $tableSwitchLabel='Rows';
  if($element=~/^table\.column/){
    $tableSwitchPath="$path.table.rows";
    $tableSwitchLabel='Columns'; 
    }
  else{$tableSwitchPath="$path.table.columns";}

  if($element eq 'table' || $element=~/^table.column/){ 
    my $attributesPath,$attributesElement;
    if($element=~/^table(\.(.*))?$/){
      $attributesPath="$path.table";
      $attributesElement=$2;
      }
    my $attributesObj=new Attributes($planner,$attributesDoc);
    my $attributesHtml=$attributesObj->html($attributesPath,$attributesElement,$action,$canEdit,$canCntl);
    $html.=$attributesHtml;
    }

  my $rowRequest='Edit';
  $rowRequest='Display' if($action eq 'Edit' && $element=~/^table.row/);
  my $iconlink1;
  my $iconlink2;
  if($vertical){
    $iconlink1="<a href='Planner.cgi?request=$rowRequest;path=$path.table.entries'><img src='planTables.gif' border=0 title='Open or close table entries'></a>";
    if($rowId){
      $iconlink2="<a href='Planner.cgi?request=$action;path=$path.table.row($rowId)'><img src='planTable.gif' border=0 title='Switch to table rows'></a>";
      }
    else{
      $iconlink2="<a href='Planner.cgi?request=$action;path=$path.table.rows'><img src='planTable.gif' border=0 title='Switch to table rows'></a>";
      }
    }
  else{
    $iconlink1="<a href='Planner.cgi?request=$rowRequest;path=$path.table.rows'><img src='planTable.gif' border=0 title='Open or close table rows'></a>";
    if($rowId){
      $iconlink2="<a href='Planner.cgi?request=$action;path=$path.table.entry($rowId)'><img src='planTables.gif' border=0 title='Switch to table entries'></a>";
      }
    else{
      $iconlink2="<a href='Planner.cgi?request=$action;path=$path.table.entries'><img src='planTables.gif' border=0 title='Switch to table entries'></a>";
      }
    }
  $html.="\n</td></tr><tr><th>$iconlink1&nbsp;$iconlink2&nbsp;Table Rows</th></tr><tr><td>\n";
  #if($element eq 'table' || $element=~/^table.row/){
  my $rowAction=$action;
  $rowAction='Display' if(!($element=~/^table.row/));
  $rowAction='Display' if($rowAction eq 'Edit' && !$canEdit);
  if($rowAction eq 'Edit'){  # If edit requested and authorized
    my $rowDoc;
    my $rownum=0;
    # Find high row number
    foreach my $rowD ($rowsDoc->children('row')){
      next if(!$rowD);
      my $rowNum=$rowD->attribute('id');
      $rownum=$rowNum if($rowNum>$rownum);
      $rowDoc=$rowD if($rowNum eq $rowId);
      }
    my $rowState;
    my $cell;
    if($rowId ne ''){  # if we have a rowId then get the values
      if($rowDoc){
        $rowState=$rowDoc->attribute('state');
        foreach $cell ($rowDoc->children('cell')){
          $cell{$cell->attribute('id')}=$cell->value;
          }
        }
      }
    $html.="<form style='margin-bottom:0;' name=TableRowEdit action='Planner.cgi' method=post>\n";
    $html.="<input type='hidden' name='request' value='Update'>\n";
    $html.="<input type='hidden' name='action' value=''/>\n";
    $rowId="$1" if($element=~/^table\.row\((\d+)\)$/);
    $html.="<input type=hidden name=row value='$rowId'>\n" if($rowDoc);
    if($vertical){
      $html.="<input type=hidden name=path value='$path.table.entry'>\n";
      }
    else{
      $html.="<input type=hidden name=path value='$path.table.row'>\n";
      }
    # indentation is off below - move left if it stays this way
      $html.="<table width='100%' class='fill'><tr><th>Row</th><th colspan=2>";
      $html.="<select name='state'>\n";
      foreach my $state ('enabled','disabled'){
        my $selected='';
        $selected=' SELECTED' if($state eq $rowState);
        $html.="<option value='$state'$selected>$state</option>\n";
        }
      $html.="</select>\n";
      $html.="&nbsp; <input type=image src='planAdd.gif' title='Create new entry' name='image' onclick=\"this.form.action.value='Add'\"/>\n";
      $html.="&nbsp; <input type=image src='planUpdate.gif' title='Write changes to selected entry' name='image' onclick=\"this.form.action.value='Update'\">\n" if($rowDoc); # only allow update if we found the row
      $html.="</td></tr>\n";
      $html.="<tr><th width='5%'>Attribute</th><th>Value</th><th>Comment</th><tr>\n";
      foreach my $colId (sort {$a<=>$b} keys %column){
        my $columnDoc=$column{$colId};
        my $columnName=$columnDoc->attribute('name');
        my $columnComment=$columnDoc->attribute('comment');
        my $columnTitle=$columnDoc->attribute('title');
        $columnTitle=$columnDoc->attribute('tip') if(!$columnTitle);  # support old attribute
        my $columnType=$columnDoc->attribute('type');
        my $columnValue=$cell{$colId};
        my $columnFormat=$columnDoc->attribute('format');
        $columnFormat=$columnDoc->attribute('menu') if(!$columnFormat); # support deprecated attribute
        $html.="<tr><td><b>$columnName</b></td>";
        if($columnType eq 'code'){
          $html.="<td colspan=3>$columnTitle<br><textarea name=cell$colId rows=10 cols=150 title='$columnTitle'>$columnValue</textarea></td></tr>";
          }
        else{
          $html.="<td>";
          if($columnFormat=~/;/){  # menu format
            $html.="<select name='cell$colId'>\n";
            my @option=split(/;/,$columnFormat);
            foreach my $option (sort @option){
              my $selected='';
              $selected=' SELECTED' if($option eq $columnValue);
              $html.="<option value='$option'$selected>$option</option>\n";
              }
            $html.="</select>\n";
            }
          else{$html.="<input type=text name=cell$colId value='$columnValue' size=80>";}
          $html.=" $columnTitle</td><td>$columnComment</td></tr>\n";
          }
        }
      $html.="<tr><th>Change</th><th colspan=2><input type=text name=change size=80> Short description of change.</th></tr>\n";
      $html.="<tr><th>Reason</th><th colspan=2><input type=text name=reason size=80> Change request number and/or short explanation.</th></tr>\n";
      $html.="</table>\n";
      }
    # indentation is off above

  # table.rows
  if($vertical){
    $html.="<table width='100%' border=0 cellpadding=3 cellspacing=1>\n";
    }
  else{
      $html.="<table width='100%' border=0 cellpadding=3 cellspacing=1><tr><th width='20'>\n";
      if($rowAction eq 'Edit'){
        $html.="<input type=image src='planUp.gif' name='image' onclick=\"this.form.action.value='MoveUp'\" title='Move selected rows above the first selected row.'/>";
        $html.="<input type=image src='planDown.gif' name='image' onclick=\"this.form.action.value='MoveDown'\" title='Move selected rows below the last selected row.'/>\n";
        $html.="</th><th width='37'>";
        $html.="<input type=image src='planToggle.gif' name=image onclick=\"this.form.action.value='Toggle'\" title='Toggle state of selected rows'>\n";
        $html.="<input type=image src='planCut.gif' name=image onclick=\"this.form.action.value='Delete'\" title='Delete selected rows'></th>\n";
        }
      else{$html.="</th><th width='19'></th>";}
      foreach my $id (sort {$a<=>$b} keys %column){
        $name=$column{$id}->attribute('name');
        $html.="<th>$name</th>";
        }
      $html.="<th>Revised</th><th>Planner</th><th>Reason</th><tr>\n";
      }
    my $rows=$tableDoc->child('rows');
    foreach my $row ($rows->children('row')){
      next if(!$row);
      $row{$row->attribute('id')}=$row;
      }
    my $row=1;
    foreach my $id (sort {$a<=>$b} keys %row){
      my $rowDoc=$row{$id};
      my $rowTime=$rowDoc->attribute('time');
      my $rowDate='';
      $rowDate=$planner->getUserDate($rowTime) if($rowTime);
      my $rowUserId=$rowDoc->attribute('user');
      my $rowPlanner='';
      if($rowUserId){
        my $userObj=$planner->getUser($rowUserId);
        if($userObj){$rowPlanner=$userObj->getName();}
        else{$rowPlanner=$rowUserId;}
        }
      my $rowReason=$rowDoc->attribute('reason');
      $state=$rowDoc->attribute('state');      # get the state 0=disable, 1=enabled
      if($state eq 'deleted'){
        $fontOn="<font color='#808080'>";  # should use style instead
        $fontOff='</font>';
        $stateImage='deleted.gif';
        $checkName='enable';
        }
      elsif($state eq 'disabled'){
        $fontOn="<font color='#808080'>";
        $fontOff="</font>";
        $stateImage='planDisabled.gif';
        $checkName='enable';
        }
      else{
        $fontOn="";
        $fontOff="";
        $stateImage='planEnabled.gif';
        $checkName='disable';
        }
      my $linkRequest='Display';
      $linkRequest='Edit' if($canEdit);
      my %cell;
      foreach $cell ($row{$id}->children('cell')){
        # 2011-02-27 eat - included encoding of cell value
        $cell{$cell->attribute('id')}=$planner->link($path,$planner->encode($cell->value));
        }
      if($vertical){
        $class='odd';
        $class='marker' if($id eq $rowId);
        $html.="<tr class='$class'><td width=20>$row</td><td>";
        $html.="<a href='Planner.cgi?request=$linkRequest;path=$path.table.entry($id)'><img src='$stateImage' border=0 title='Edit row'></a>";
        if($rowAction eq 'Edit'){
          $html.="<input type=checkbox name=$checkName value=$id>" if($rowAction eq 'Edit');
          $html.="<input type=image src='planUp.gif' name='image' onclick=\"this.form.action.value='MoveUp'\" title='Move selected rows above the first selected row.'/>";
          $html.="<input type=image src='planDown.gif' name='image' onclick=\"this.form.action.value='MoveDown'\" title='Move selected rows below the last selected row.'/>\n";
          $html.="<input type=image src='planToggle.gif' name='image' onclick=\"this.form.action.value='Toggle'\" title='Toggle state of selected rows'>\n";
          $html.="<input type=image src='planCut.gif' name='image' onclick=\"this.form.action.value='Delete'\" title='Delete selected rows'>\n";
          }
        $html.="</td></tr>\n";
        my $cellrow=0;
        foreach $column (sort {$a<=>$b} keys %column){
          $class='even';
          $class='odd' if($cellrow%2);
          $value=$cell{$column};
          $value=$column{$column}->attribute('default') if(!defined $value);
          $type=$column{$column}->attribute('type');
          my $name=$column{$column}->attribute('name');
          $html.="<tr class='$class'><td><b>$name</b></td><td>";
          if($type eq 'code'){          
            $html.="<pre>$fontOn$value$fontOff</pre>";
            }
          elsif($column{$column}->attribute('for') ne 'comment'){          
            $html.="<a href='nbe.cgi?request=Pivot;view=.generic;keyValue=$value;form=on;'>$fontOn$value$fontOff</a>";
            }
          else{$html.="$fontOn$value$fontOff";}
          $html.="</td></tr>\n";
          $cellrow++;
          }
        $html.="<tr><td class='fill'><b>Revised</b></td><td class=fill>$rowDate</td></tr>\n";
        $html.="<tr><td class='fill'><b>Planner</b></td><td class=fill>$rowPlanner</td></tr>\n";
        $html.="<tr><td class='fill'><b>Reason</b></td><td class=fill>$rowReason</td></tr>\n";
        }
      else{
        $class='even';
        $class='odd' if($row%2);
        $class='marker' if($id eq $rowId);
        $html.="<tr class=$class>";
        $html.="<td>$row</td>\n<td><a href='Planner.cgi?request=$linkRequest;path=$path.table.row($id)'><img src='$stateImage' border=0 title='Edit row'></a>";
        $html.="<input type=checkbox name=$checkName value=$id>" if($rowAction eq 'Edit');
        $html.="</td>";
        foreach $column (sort {$a<=>$b} keys %column){
          $value=$cell{$column};
          $value=$column{$column}->attribute('default') if(!defined $value);
          $type=$column{$column}->attribute('type');
          if($type eq 'code'){          
            $html.="<td><pre>$fontOn$value$fontOff</pre></td>";
            }
          elsif($column{$column}->attribute('for') ne 'comment'){          
            $html.="<td><a href='nbe.cgi?request=Pivot;view=.generic;keyValue=$value;form=on;'>$fontOn$value$fontOff</a></td>";
            }
          else{$html.="<td>$fontOn$value$fontOff</td>";}
          }
        $html.="<td class=fill>$rowDate</td><td class=fill>$rowPlanner</td><td class=fill>$rowReason</td></tr>\n";
        }
      $row++;
      }
    $html.="</table>\n";
    $html.="</form>" if($rowAction eq 'Edit');
   #}
  $html.="</td></tr></table>\n";
  return($html);
  }

# Generate CFG file for text editors

sub cfg{
  my ($self)=@_;
  my $tableDoc=$self->{DOC};
  my $planner=$self->{PLANNER};
  return("# NBK000W Table element not found\n\n# Attributes:\n\n# Rows:\n\n") if(!$tableDoc);
  my $attributesDoc=$tableDoc->child('attributes');
  $attributesDoc=$tableDoc->child('columns') if(!$attributesDoc);
  return("# NBK000W Columns element not found\n\n# Attributes:\n\n# Rows:\n\n") if(!$attributesDoc);
  my $nb='';

  $nb.=")table\n";
  # display attributes using Attributes.pm
  my $attributesObj=new Attributes($planner,$attributesDoc);
  $nb.=$attributesObj->cfg();

  my $rowsDoc=$tableDoc->child('rows');
  return("# NBK000W Rows element not found\n\n# Rows:\n\n") if(!$rowsDoc);

  my %column;
  # support both new and old element names
  foreach my $elementName ('attribute','column'){
    foreach my $element ($attributesDoc->children($elementName)){
      next if(!$element);
      my $columnState=$element->attribute('state');
      #my $columnFor=$element->attribute('for');
      $columnElement{$element->attribute('id')}=$element;
      }
    }

  undef %row;
  foreach my $row ($rowsDoc->children('row')){
    next if(!$row);
    $rowId=$row->attribute('id');
    $rowState=$row->attribute('state');
    my $stateFlag='';
    $stateFlag='!' if($rowState eq 'disabled');
    $nb.="]$stateFlag\n";
    undef %cellValue;
    $comment='';
    foreach $cell ($row->children('cell')){
      $cellValue{$cell->attribute('id')}=$cell->value;
      }
    foreach my $columnId (sort {$a<=>$b} keys %columnElement){
      my $element=$columnElement{$columnId};
      my $name=$element->attribute('name');
      my $value=$cellValue{$columnId};
      $value=$planner->decode($value);
      if($value){
        $nb.=" $name";
        my @text=split(/\r?\n/,$value);
        if($#text>0){
          $nb.="{\n";
          foreach my $text (@text){
            $nb.=" $text\n";
            }
          $nb.='}';
          }
        else{$nb.=": $value";}
        $nb.="\n";
        }
      }
    }
  return($nb);
  }

sub table{
  my ($self)=@_;
  my $tableDoc=$self->{DOC};
  my $planner=$self->{PLANNER};
  return("# NBK000W Table element not found\n") if(!$tableDoc);
  my $attributesDoc=$tableDoc->child('attributes');
  $attributesDoc=$tableDoc->child('columns') if(!$attributesDoc);
  return("# NBK000W Columns element not found\n") if(!$attributesDoc);
  my $rowsDoc=$tableDoc->child('rows');
  return("# NBK000W Rows element not found\n\n# Rows:\n\n") if(!$rowsDoc);

  my %column;
  foreach my $elementName ('attribute','column'){
    foreach my $element ($attributesDoc->children($elementName)){
      next if(!$element);
      my $columnState=$element->attribute('state');
      my $columnFor=$element->attribute('for');
      $columnElement{$element->attribute('id')}=$element if($columnFor ne 'opt');
      }
    }
  my $table=']';
  foreach my $columnId (sort {$a<=>$b} keys %columnElement){
    my $element=$columnElement{$columnId};
    my $name=$element->attribute('name');
    $table.=";$name";
    }
  $table.="\n";

  my %row;
  foreach my $row ($rowsDoc->children('row')){
    next if(!$row);
    $rowId=$row->attribute('id');
    $rowState=$row->attribute('state');
    my $stateFlag='1';
    $stateFlag='0' if($rowState eq 'disabled');
    $table.="$stateFlag";
    my %cellValue;
    $comment='';
    foreach my $cell ($row->children('cell')){
      $cellValue{$cell->attribute('id')}=$cell->value;
      }
    foreach my $columnId (sort {$a<=>$b} keys %columnElement){
      my $element=$columnElement{$columnId};
      my $value=$cellValue{$columnId};
      $value=$planner->decode($value);
      $table.=";$value";
      }
    $table.="\n";
    }
  return($table);
  }

#  Extract option values and enabled table rows using selected attributes
#
#    $msg=extract(\%option,\@table,"attribute,attribute,attribute,...");
#    if($msg) ... error ...
#    $attributeValue=$option{<attribute>};
#    foreach my @row (@table){
#      my ($attr1,$attr2,...)=@row;
#      # process row
#      }
#
sub extract{
  my ($self,$option,$table,$attributes)=@_;
  my $tableDoc=$self->{DOC};
  my $planner=$self->{PLANNER};
  return("CAB000W Table element not found\n") if(!$tableDoc);
  my $attributesDoc=$tableDoc->child('attributes');
  $attributesDoc=$tableDoc->child('columns') if(!$attributesDoc);
  return("CAB000W Columns element not found\n") if(!$attributesDoc);
  my $rowsDoc=$tableDoc->child('rows');
  return("CAB000W Rows element not found\n") if(!$rowsDoc);

  my $colSeq;
  my %colSeq;
  my $colNum=0;
  my $state=0;
  foreach my $colName (split(/,/,$attributes)){
    $state=1 if($colName eq '_state'); # user wants row state returned
    $colNum++;
    $colSeq{$colName}=$colNum;
    }
  my %column;
  my %colNum;
  my %colValue;
  foreach my $elementName ('attribute','column'){
    foreach my $element ($attributesDoc->children($elementName)){
      next if(!$element);
      next if($element->attribute('state') eq 'disabled');   # ignore disabled attributes
      my $colName=$element->attribute('name');
      my $columnFor=$element->attribute('for');
      if($columnFor eq 'opt'){
        my $value=$element->value;
        $value=$element->attribute('value') if(!$value); # old style
        $value=$1 if($value=~/^\r?\n(.*)$/s);  # remove first leading new line
        $$option{$colName}=$value;
        }
      else{
        $colSeq=$colSeq{$colName};
        next if(!$colSeq);
        $colNum{$element->attribute('id')}=$colSeq;
        $colValue{$colSeq}=$element->value;
        $colValue{$colSeq}=$element->attribute('value') if($element->attribute('value'));      # old method
        $colValue{$colSeq}=$element->attribute('default') if($element->attribute('default'));  # old method
        }
      }
    }
  my %row;
  foreach my $row ($rowsDoc->children('row')){
    next if(!$row);
    next if(!$state && $row->attribute('state') eq 'disabled');
    my %cellValue;
    for(my $i=1;$i<=$colNum;$i++){                           # initialize cells to default values
      $cellValue{$i}=$colValue{$i};
      }
    $cellValue{$colSeq{'_state'}}=$row->attribute('state') if($state);
    foreach my $cell ($row->children('cell')){              # get values from row
      my $cellId=$cell->attribute('id');
      $cellValue{$colNum{$cellId}}=$planner->decode($cell->value) if($colNum{$cellId});
      }
    my @row;
    foreach my $colNum (sort {$a<=>$b} keys %cellValue){    # convert row to list
      push(@row,$cellValue{$colNum});
      }
    push(@$table,\@row);
    }
  return('');
  }

1;
